@@ -4,8 +4,8 @@
 
 ;; Author: Le Wang <lewang.dev.26@gmail.com>
 ;; URL: https://github.com/lewang/spatial-window
-;; Package-Version: 0.9.3
-;; Package-Revision: 5388bb6c1989
+;; Package-Version: 20260215.304
+;; Package-Revision: d194ae2864c3
 ;; Package-Requires: ((emacs "28.1") (posframe "1.0.0"))
 ;; Keywords: convenience, windows
 
@@ -133,6 +133,73 @@ Returns alist of (extension-key . base-key)."
     ('colemak spatial-window-extensions-colemak)
     (_ nil)))
 
+;;; Frame introspection
+
+(defun spatial-window--window-bounds (&optional frame)
+  "Return window bounds for FRAME as list of (window x-start x-end y-start y-end).
+Coordinates are normalized to 0.0-1.0 range relative to frame size."
+  (let* ((frame (or frame (selected-frame)))
+         (frame-w (float (frame-pixel-width frame)))
+         (frame-h (float (frame-pixel-height frame)))
+         (windows (window-list frame 'no-minibuf)))
+    (mapcar (lambda (w)
+              (let ((edges (window-pixel-edges w)))
+                (list w
+                      (/ (nth 0 edges) frame-w)   ; x-start
+                      (/ (nth 2 edges) frame-w)   ; x-end
+                      (/ (nth 1 edges) frame-h)   ; y-start
+                      (/ (nth 3 edges) frame-h)))) ; y-end
+            windows)))
+
+;;; Layout-dependent key mapping
+
+(defun spatial-window--final-to-keys (final kbd-rows kbd-cols kbd-layout)
+  "Convert FINAL assignment grid to alist of (window . keys).
+KBD-ROWS and KBD-COLS define grid dimensions, KBD-LAYOUT maps to key strings."
+  (let ((result (make-hash-table :test 'eq)))
+    (dotimes (row kbd-rows)
+      (dotimes (col kbd-cols)
+        (let ((win (aref (aref final row) col)))
+          (when win
+            (push (nth col (nth row kbd-layout)) (gethash win result))))))
+    ;; Convert to alist with reversed key lists
+    (let ((alist nil))
+      (maphash (lambda (win keys)
+                 (push (cons win (nreverse keys)) alist))
+               result)
+      alist)))
+
+(defun spatial-window--assignment-to-keys (grid kbd-layout)
+  "Map assignment GRID to keyboard keys using KBD-LAYOUT.
+GRID is the 2D vector returned by `spatial-window--compute-assignment'.
+KBD-LAYOUT is the keyboard layout (list of rows of key strings).
+Returns alist of (window . (list of keys)), or nil if layout is invalid."
+  (let ((lengths (mapcar #'length kbd-layout)))
+    (cond
+     ((or (null kbd-layout) (memq 0 lengths))
+      (message "Invalid keyboard layout: empty layout or empty row")
+      nil)
+     ((not (apply #'= lengths))
+      (message "Invalid keyboard layout: rows have different lengths")
+      nil)
+     (t
+      (spatial-window--final-to-keys grid (length kbd-layout) (car lengths) kbd-layout)))))
+
+(defun spatial-window--format-key-grid (keys kbd-layout)
+  "Format KEYS as a keyboard grid string using KBD-LAYOUT.
+Returns a string showing which keys are assigned, displayed in keyboard layout."
+  (let ((key-set (make-hash-table :test 'equal)))
+    (dolist (k keys)
+      (puthash k t key-set))
+    (mapconcat
+     (lambda (row)
+       (mapconcat
+        (lambda (key)
+          (if (gethash key key-set) key "·"))
+        row " "))
+     kbd-layout
+     "\n")))
+
 (defface spatial-window-overlay-face
   '((t (:foreground "red" :background "white" :weight bold)))
   "Face for spatial-window key overlay."
@@ -150,7 +217,8 @@ Returns alist of (extension-key . base-key)."
   action
   overlay-timer
   history-cursor
-  history-live-config)
+  history-live-config
+  kill-soft-select)
 
 (defvar spatial-window--state (spatial-window--make-state)
   "Active session state for spatial-window.")
@@ -178,7 +246,6 @@ If SELECTED-P, use selected face with border."
 Reads assignments from state (must be set by caller).
 If SELECTED-WINDOWS is non-nil, highlight those windows with a border.
 Returns non-nil if overlays were shown, nil if no assignments."
-  (require 'posframe)
   ;; Clean up any existing posframes first
   (spatial-window--remove-overlays)
   (let* ((st spatial-window--state)
@@ -189,7 +256,7 @@ Returns non-nil if overlays were shown, nil if no assignments."
       (dolist (pair assignments)
         (let* ((window (car pair))
                (keys (cdr pair))
-               (grid-str (spatial-window--format-key-grid keys))
+               (grid-str (spatial-window--format-key-grid keys (spatial-window--get-layout)))
                (buf-name (format " *spatial-window-%d*" idx))
                (edges (window-pixel-edges window))
                (x (nth 0 edges))
@@ -251,26 +318,25 @@ If no target window found (ambiguous key), do nothing."
        ,@body)))
 
 (defun spatial-window--exit-selection-mode ()
-  "Exit selection mode: cancel timer, clear flag, and remove overlays."
-  (let ((timer (spatial-window--state-overlay-timer spatial-window--state)))
-    (when (timerp timer)
-      (cancel-timer timer)))
-  (setf (spatial-window--state-selection-active spatial-window--state) nil)
-  (spatial-window--remove-overlays))
-
-(defun spatial-window--abort ()
-  "Abort window selection and clean up overlays."
-  (interactive)
+  "Exit selection mode: cancel timer, clear flag, and remove overlays.
+If the user accepted a historical layout via ←/→ navigation, save the
+pre-browsing layout onto the history ring so the navigation is undoable."
   (let* ((st spatial-window--state)
          (timer (spatial-window--state-overlay-timer st))
          (live (spatial-window--state-history-live-config st)))
     (when (timerp timer)
       (cancel-timer timer))
     (when live
-      (set-window-configuration live))
-    (spatial-window--remove-overlays)
-    (spatial-window--reset-state)
-    (keyboard-quit)))
+      (spatial-window--save-layout 'undo live)
+      (setf (spatial-window--state-history-live-config st) nil))
+    (setf (spatial-window--state-selection-active st) nil)
+    (spatial-window--remove-overlays)))
+
+(defun spatial-window--abort ()
+  "Abort window selection and clean up overlays."
+  (interactive)
+  (spatial-window--cleanup-mode)
+  (keyboard-quit))
 
 (defun spatial-window--reset-state ()
   "Reset all state variables for action modes."
@@ -292,12 +358,18 @@ If no target window found (ambiguous key), do nothing."
     (select-window (minibuffer-window))))
 
 (defun spatial-window--cleanup-mode ()
-  "Clean up overlays, cancel timers, and reset state after any mode ends."
-  (let ((timer (spatial-window--state-overlay-timer spatial-window--state)))
+  "Clean up overlays, cancel timers, and reset state after any mode ends.
+If the user was browsing history and the transient map exits
+unexpectedly (e.g. unbound key), restore the pre-browsing layout."
+  (let* ((st spatial-window--state)
+         (timer (spatial-window--state-overlay-timer st))
+         (live (spatial-window--state-history-live-config st)))
     (when (timerp timer)
-      (cancel-timer timer)))
-  (spatial-window--remove-overlays)
-  (spatial-window--reset-state))
+      (cancel-timer timer))
+    (when live
+      (set-window-configuration live))
+    (spatial-window--remove-overlays)
+    (spatial-window--reset-state)))
 
 (defun spatial-window--make-mode-keymap (key-action &optional extra-bindings)
   "Create keymap binding all layout keys to KEY-ACTION.
@@ -319,8 +391,12 @@ EXTRA-BINDINGS is an alist of (key-string . command) for additional bindings.
 KEYMAP is the transient keymap to activate.
 HIGHLIGHTED is a list of windows to highlight in overlays.
 MESSAGE is displayed in the minibuffer."
+  (require 'posframe)
   (let ((st spatial-window--state))
-    (setf (spatial-window--state-assignments st) (spatial-window--assign-keys)
+    (setf (spatial-window--state-assignments st)
+          (spatial-window--assignment-to-keys
+           (spatial-window--compute-assignment (spatial-window--window-bounds))
+           (spatial-window--get-layout))
           (spatial-window--state-highlighted-windows st) highlighted)
     (when (spatial-window--state-assignments st)
       (setf (spatial-window--state-selection-active st) t)
@@ -341,8 +417,8 @@ MESSAGE is displayed in the minibuffer."
                 (spatial-window--state-selection-active spatial-window--state))))
        #'spatial-window--cleanup-mode))))
 
-(defun spatial-window--kill-multi-mode-message ()
-  "Display kill-multi mode status message."
+(defun spatial-window--kill-mode-message ()
+  "Display kill mode status message."
   (let ((n (length (spatial-window--state-selected-windows spatial-window--state))))
     (message "RET to kill %d window%s. C-g to abort."
              n (if (= n 1) "" "s"))))
@@ -388,129 +464,127 @@ Oldest entries are evicted when this limit is exceeded."
         (tab-bar-tabs-set tabs))
     (set-frame-parameter nil 'spatial-window-config history)))
 
-(defun spatial-window--save-layout (action)
-  "Push current window configuration onto history with ACTION tag.
+(defun spatial-window--save-layout (action &optional config)
+  "Push window configuration onto history with ACTION tag.
+Uses CONFIG if provided, otherwise `current-window-configuration'.
 Evicts oldest entry when `spatial-window-history-max' is exceeded."
-  (let* ((entry (cons action (current-window-configuration)))
+  (let* ((entry (cons action (or config (current-window-configuration))))
          (history (cons entry (spatial-window--get-history))))
     (when (> (length history) spatial-window-history-max)
       (setcdr (nthcdr (1- spatial-window-history-max) history) nil))
     (spatial-window--set-history history)))
 
-(defun spatial-window--restore-layout ()
-  "Pop and restore the most recent window configuration.
-Each entry is (ACTION . WINDOW-CONFIGURATION).
-Returns the ACTION symbol on success, nil otherwise."
-  (let ((history (spatial-window--get-history)))
-    (when history
-      (let ((entry (car history)))
-        (set-window-configuration (cdr entry))
-        (spatial-window--set-history (cdr history))
-        (car entry)))))
-
-
-(defun spatial-window-unfocus ()
-  "Restore the previously saved window layout."
-  (interactive)
-  (if (spatial-window--restore-layout)
-      (message "Restored window layout")
-    (beep)
-    (message "No saved layout to restore")))
-
 ;;; Unified selection with action modifiers
 
+(defun spatial-window--refresh-overlays ()
+  "Refresh overlays if currently visible."
+  (let ((st spatial-window--state))
+    (when (spatial-window--state-overlays-visible st)
+      (spatial-window--show-overlays (spatial-window--state-highlighted-windows st)))))
+
+(defun spatial-window--complete-single-input (win)
+  "Exit selection mode and complete current single-input action on WIN."
+  (spatial-window--exit-selection-mode)
+  (pcase (spatial-window--state-action spatial-window--state)
+    ('swap
+     (spatial-window--save-layout 'swap)
+     (spatial-window--swap-windows
+      (spatial-window--state-source-window spatial-window--state) win)
+     (select-window win)
+     (message "Swapped windows"))
+    ('focus
+     (spatial-window--save-layout 'focus)
+     (select-window win)
+     (let ((ignore-window-parameters t))
+       (delete-other-windows win))
+     (message "Focused window"))
+    (_ (select-window win))))
+
+(defun spatial-window--set-action (action message)
+  "Switch to ACTION mode, highlight current window, show MESSAGE.
+Callers should set additional state fields before calling this."
+  (let ((st spatial-window--state))
+    (setf (spatial-window--state-action st) action
+          (spatial-window--state-highlighted-windows st) (list (selected-window)))
+    (spatial-window--refresh-overlays))
+  (message "%s" message))
+
 (defun spatial-window--act-by-key ()
   "Apply current action to the window corresponding to the pressed key."
   (interactive)
   (spatial-window--with-target-window
-    (let ((action (spatial-window--state-action spatial-window--state)))
-      (pcase action
-        ('multi-kill
-         (let ((st spatial-window--state))
-           (if (memq win (spatial-window--state-selected-windows st))
-               (setf (spatial-window--state-selected-windows st)
-                     (delq win (spatial-window--state-selected-windows st)))
-             (push win (spatial-window--state-selected-windows st)))
-           (setf (spatial-window--state-highlighted-windows st)
-                 (spatial-window--state-selected-windows st))
-           (when (spatial-window--state-overlays-visible st)
-             (spatial-window--show-overlays (spatial-window--state-highlighted-windows st)))
-           (spatial-window--kill-multi-mode-message)))
-        (_
-         (spatial-window--exit-selection-mode)
-         (pcase action
-           ('kill
-            (when (window-live-p win)
-              (spatial-window--save-layout 'kill)
-              (delete-window win))
-            (message "Killed window"))
-           ('swap
-            (spatial-window--save-layout 'swap)
-            (spatial-window--swap-windows
-             (spatial-window--state-source-window spatial-window--state) win)
-            (select-window win)
-            (message "Swapped windows"))
-           ('focus
-            (spatial-window--save-layout 'focus)
-            (select-window win)
-            (let ((ignore-window-parameters t))
-              (delete-other-windows win))
-            (message "Focused window (unfocus to restore)"))
-           (_
-            (select-window win))))))))
+    (pcase (spatial-window--state-action spatial-window--state)
+      ;; Multi-input: toggle window in selection set
+      ('kill
+       (let* ((st spatial-window--state)
+              (soft (spatial-window--state-kill-soft-select st)))
+         ;; First press on a different window replaces the soft pre-selection
+         (when (and soft (not (eq win soft)))
+           (setf (spatial-window--state-selected-windows st)
+                 (delq soft (spatial-window--state-selected-windows st))))
+         (when soft
+           (setf (spatial-window--state-kill-soft-select st) nil))
+         (if (memq win (spatial-window--state-selected-windows st))
+             (setf (spatial-window--state-selected-windows st)
+                   (delq win (spatial-window--state-selected-windows st)))
+           (push win (spatial-window--state-selected-windows st)))
+         (setf (spatial-window--state-highlighted-windows st)
+               (spatial-window--state-selected-windows st))
+         (spatial-window--refresh-overlays)
+         (spatial-window--kill-mode-message)))
+      ;; Single-input: immediately complete
+      (_ (spatial-window--complete-single-input win)))))
 
 (defun spatial-window--set-action-kill ()
-  "Switch to kill action."
-  (interactive)
-  (setf (spatial-window--state-action spatial-window--state) 'kill)
-  (message "KILL: select window to kill"))
-
-(defun spatial-window--set-action-multi-kill ()
-  "Switch to multi-kill action."
+  "Switch to kill action with current window soft-pre-selected.
+The pre-selection is replaced if the first key press picks a
+different window, avoiding an extra deselect step."
   (interactive)
   (let ((st spatial-window--state))
-    (setf (spatial-window--state-action st) 'multi-kill
-          (spatial-window--state-selected-windows st) nil))
-  (message "MULTI-KILL: toggle windows, RET to kill"))
+    (setf (spatial-window--state-selected-windows st) (list (selected-window))
+          (spatial-window--state-kill-soft-select st) (selected-window)))
+  (spatial-window--set-action 'kill "KILL: toggle windows, RET to delete"))
 
-(defun spatial-window--execute-multi-kill-unified ()
-  "Execute multi-kill in unified mode."
+(defun spatial-window--execute-ret ()
+  "Execute RET: confirm current action on selected/current window."
   (interactive)
-  (if (not (eq (spatial-window--state-action spatial-window--state) 'multi-kill))
-      (spatial-window--exit-selection-mode)
-    (let ((windows-to-kill (spatial-window--state-selected-windows spatial-window--state)))
-      (spatial-window--exit-selection-mode)
-      (when windows-to-kill
-        (spatial-window--save-layout 'kill)
-        (dolist (win windows-to-kill)
-          (when (window-live-p win)
-            (delete-window win))))
-      (message "Killed %d window(s)" (length windows-to-kill)))))
+  (pcase (spatial-window--state-action spatial-window--state)
+    ;; Multi-input: confirm and execute on accumulated set
+    ('kill
+     (let ((windows-to-kill (spatial-window--state-selected-windows spatial-window--state)))
+       (spatial-window--exit-selection-mode)
+       (when windows-to-kill
+         (spatial-window--save-layout 'kill)
+         (dolist (win windows-to-kill)
+           (when (window-live-p win)
+             (delete-window win))))
+       (message "Killed %d window(s)" (length windows-to-kill))))
+    ;; Single-input: complete on current window
+    (_ (spatial-window--complete-single-input (selected-window)))))
 
 (defun spatial-window--set-action-swap ()
   "Switch to swap action, recording current window as source."
   (interactive)
-  (let ((st spatial-window--state))
-    (setf (spatial-window--state-action st) 'swap
-          (spatial-window--state-source-window st) (selected-window)
-          (spatial-window--state-highlighted-windows st)
-          (list (selected-window)))
-    (when (spatial-window--state-overlays-visible st)
-      (spatial-window--show-overlays (spatial-window--state-highlighted-windows st))))
-  (message "SWAP: select target window"))
+  (setf (spatial-window--state-source-window spatial-window--state) (selected-window))
+  (spatial-window--set-action 'swap "SWAP: select target window"))
 
 (defun spatial-window--set-action-focus ()
-  "Switch to focus action."
+  "Switch to focus action with current window highlighted."
   (interactive)
-  (setf (spatial-window--state-action spatial-window--state) 'focus)
-  (message "FOCUS: select window to focus"))
+  (spatial-window--set-action 'focus "FOCUS: select window, RET to focus current"))
 
 (defun spatial-window--history-refresh ()
-  "Recompute assignments and refresh overlays after history navigation."
+  "Recompute assignments and refresh overlays after history navigation.
+Update highlighted window to match `selected-window' from the
+restored configuration so the red highlight tracks the cursor."
   (let ((st spatial-window--state))
-    (setf (spatial-window--state-assignments st) (spatial-window--assign-keys))
-    (when (spatial-window--state-overlays-visible st)
-      (spatial-window--show-overlays (spatial-window--state-highlighted-windows st)))
+    (setf (spatial-window--state-highlighted-windows st)
+          (list (selected-window))
+          (spatial-window--state-assignments st)
+          (spatial-window--assignment-to-keys
+           (spatial-window--compute-assignment (spatial-window--window-bounds))
+           (spatial-window--get-layout)))
+    (spatial-window--refresh-overlays)
     (message "%s" (spatial-window--unified-mode-message))))
 
 (defun spatial-window--history-back ()
@@ -584,7 +658,7 @@ While browsing: shows available directions and position."
 
 (defun spatial-window--unified-mode-message ()
   "Return hint message for unified selection mode."
-  (format "Select window or: [K]ill [M]ulti-kill [S]wap [F]ocus [RET]done%s"
+  (format "Select window or: [K]ill [S]wap [F]ocus [RET]done%s"
           (spatial-window--history-message-part)))
 
 (defun spatial-window--make-unified-keymap ()
@@ -592,9 +666,8 @@ While browsing: shows available directions and position."
   (let ((map (spatial-window--make-mode-keymap
               #'spatial-window--act-by-key
               '(("SPC" . spatial-window--select-minibuffer)
-                ("RET" . spatial-window--execute-multi-kill-unified)))))
+                ("RET" . spatial-window--execute-ret)))))
     (define-key map (kbd "K") #'spatial-window--set-action-kill)
-    (define-key map (kbd "M") #'spatial-window--set-action-multi-kill)
     (define-key map (kbd "S") #'spatial-window--set-action-swap)
     (define-key map (kbd "F") #'spatial-window--set-action-focus)
     (define-key map (kbd "<left>") #'spatial-window--history-back)
@@ -607,24 +680,27 @@ While browsing: shows available directions and position."
 
 Press a layout key to switch to that window immediately.
 
-Uppercase modifiers change the action before selecting:
-  K - Kill: delete a window
-  M - Multi-kill: toggle windows on/off, RET to delete all selected
+The current window is always highlighted on entry so you can see
+where you are.  Uppercase modifiers change the action:
+  K - Kill: current window pre-selected for deletion.  Press RET
+      to kill it immediately, or select a different window to
+      replace the pre-selection.  Toggle additional windows to
+      build a multi-kill set, then RET to delete all.
   S - Swap: swap buffers with current window
-  F - Focus: zoom into a window (delete others)
+  F - Focus: press RET to focus current window, or select another
 
 Other keys:
   Left/Right - Browse window configuration history
-  RET - Exit selection (execute kills in multi-kill mode)
+  RET - Confirm current selection (kill/focus) or exit
   SPC - Select minibuffer window (if active), otherwise exit
-  C-g - Abort (restores layout if browsing history)
+  \\[keyboard-quit] - Abort (restores layout if browsing history)
 
 When `spatial-window-overlay-delay' is set, overlays appear after
 the configured delay instead of immediately."
   (interactive)
   (spatial-window--setup-transient-mode
    (spatial-window--make-unified-keymap)
-   nil
+   (list (selected-window))
    (spatial-window--unified-mode-message)))
 
 (provide 'spatial-window)
